/* a state aciton advisoing behavior that picks a set of states to advise about each time */

#include <algorithm>
#include "OptQCoachSABehavior.hpp"
#include "Logger.hpp"
#include "CoachParam.hpp"
#include "random.hpp"

using namespace spades;
using namespace spadesalib;
using namespace std;

OptQCoachSABehavior::OptQCoachSABehavior(AgentCommInterface* pacs,
					 float gamma,
					 explore_t exp, float alpha_explore, float boltzmann_constant,
					 const char* enable_map_fn,
					 const char* opt_qtable_fn)
  : BaseCoachSABehavior(pacs, gamma, exp, alpha_explore, boltzmann_constant, enable_map_fn),
    qt_opt(opt_qtable_fn),
    vec_state_value(State::getNumStates()), vec_agent_response(State::getNumStates())
{
  vec_state_value_t::iterator iter;
  unsigned idx;
  for (iter = vec_state_value.begin(), idx = 0;
       iter != vec_state_value.end();
       iter++, idx++)
    {
      *iter = qt_opt.getV(idx);
    }
}

OptQCoachSABehavior::~OptQCoachSABehavior()
{
}

  
bool
OptQCoachSABehavior::initialize()
{
  return true;
}

void
OptQCoachSABehavior::respondSenseSA(Behavior::SenseType st)
{
  switch (st)
    {
    case ST_None:
      errorlog << "I don't know what to do with ST_None" << ende;
      break;
    case ST_TimeNotify:
      //do nothing, but we should never get here
      break;
    case ST_State:
      if (CoachParam::instance()->getNumPred() > 1)
	errorlog << "There's a big hack in OptQCoachSABehavior to look at pred 0" << ende;
      updateFromAgentAction(0);
      checkForQUpdate(0);
      break;
    case ST_Comm:
      errorlog << "OptQCoachSABehavior: I don't handle comms: " << getCommType() << ende;
      break;
    default:
      errorlog << "I don't understand st: " << st << ende;
      break;
    }
}


class ltMeanSummary
{
public:
  bool operator()(const MeanSummary& m1, const MeanSummary& m2)
  {
    if (m1.getNumPoints() > 0)
      {
	if (m2.getNumPoints() > 0)
	  return m1.getMean() < m2.getMean();
	return true;
      }
    else if (m2.getNumPoints() > 0)
      {
	return false;
      }
    else
      //they're equal!
      return true;
  }
};

//this is the function which is called when it is time for more advice
void
OptQCoachSABehavior::getStateActionAdvice(int max_num_to_send, vec_sa_advice_t& v)
{
  ltMeanSummary lt;
  for (int i=0; i<max_num_to_send; i++)
    {
      vec_agent_response_t::iterator iter = min_element(vec_agent_response.begin(),
							vec_agent_response.end(),
							lt);
      unsigned sidx = iter - vec_agent_response.begin();

      if (iter->getNumPoints() == 0)
	{
	  if (CoachParam::instance()->getSAOptQUseLeftover())
	    {
	      actionlog(50) << "OptQCoachSABehavior: no more states seen to advise, picking random" << ende;
	      sidx = int_random(State::getNumStates());
	    }
	  else
	    {
	      actionlog(50) << "OptQCoachSABehavior: no more states seen to advise" << ende;
	      break;
	    }
	}
      else if (iter->getMean() > -CoachParam::instance()->getSAOptQNoAdviseThreshold())
	{
	  if (CoachParam::instance()->getSAOptQUseLeftover())
	    {
	      actionlog(50) << "OptQCoachSABehavior: I can't really help on any more states: "
			    << sidx << ' ' << *iter
			    << ", picking random"
			    << ende;
	      sidx = int_random(State::getNumStates());
	    }
	  else
	    {
	      actionlog(50) << "OptQCoachSABehavior: I can't really help on any more states: "
			    << sidx << ' ' << *iter << ende;
	      break;
	    }
	}
      
      actionlog(60) << "OptQCoachSABehavior: advising on "
		    << sidx << ' ' << *iter << ende;
      v.push_back(SAAdvice(sidx,
			   (basic_move_t)qtable.getAct(exp_type, sidx, alpha_explore, boltzmann_constant)));
      iter->reset();
    }
}


void
OptQCoachSABehavior::updateFromAgentAction(entityid_t id)
{
  if (!isLastValid())
    {
      actionlog(100) << "OptQCoachSABehavior: last not valid, ignoring" << ende;
      return;
    }
  
  map_action_recv_t::const_iterator iter = getMapActionRecv().find(id);
  if (iter == getMapActionRecv().end())
    {
      errorlog << "OptQCoachSABehavior: I need to see the agent action!" << ende;
      return;
    }

  unsigned sidx = getLastState().flatVal();

  if (sidx >= vec_agent_response.size() || sidx >= vec_state_value.size())
    errorlog << "OptQCoachSABehavior::updateFromAgentAction: sidx too big: "
	     << sidx << ' '
	     << vec_agent_response.size() << ' '
	     << vec_state_value.size() << ' '
	     << ende;
  vec_agent_response[sidx].addPoint(qt_opt.getQ(sidx, iter->second) - vec_state_value[sidx]);
}

void
OptQCoachSABehavior::checkForQUpdate(entityid_t id)
{
  if (!isLastValid())
    {
      actionlog(70) << "OptQCoachSABehavior: not q updating because last not valid" << ende;
      return;
    }
  
  map_action_recv_t::const_iterator map_iter = getMapActionRecv().find(id);
  if (map_iter == getMapActionRecv().end())
    {
      errorlog << "OptQCoachSABehavior: I need to see the agent action!" << ende;
      return;
    }

  setLastAction(map_iter->second);
  
  unsigned sidx = getLastState().flatVal();

  /* Perhaps we should do the q-update for any enabled action, but it's just not clear to me
     right now */
  SAAdviceRepository::state_entry_t::const_iterator entry_iter =
    find(getAdviceRepository()->getStateAdvice(sidx).begin(),
	 getAdviceRepository()->getStateAdvice(sidx).end(),
	 getLastAction());

  if (entry_iter == getAdviceRepository()->getStateAdvice(sidx).end())
    {
      actionlog(70) << "OptQCoachSABehavior: not q updating because action not advised "
		    << sidx << ' ' << getLastAction()
		    << ende;
    }
  else
    {
      actionlog(70) << "OptQCoachSABehavior: q updating "
		    << sidx << ' ' << getLastAction()
		    << ende;
      updateQValue();
    }
}

