/* This is a virtual base class behavior for qlearn style behaviors
   it sets up some logging and other recording, as well as storing
   actions and such */

#ifndef _BASE_QLEARN_BEHAVIOR_H_
#define _BASE_QLEARN_BEHAVIOR_H_

#include <string>
#include <fstream>
#include "Behavior.hpp"
#include "QTableFlat.hpp"
#include "ActionSubsetMap.hpp"

//The behavior takes over responsibility for the memory here
class BaseQLearnBehavior : public Behavior
{
public:
  BaseQLearnBehavior(spadesalib::AgentCommInterface* pacs,
		     float gamma,
		     explore_t exp, float alpha_explore, float boltzmann_constant,
		     const char* disable_map_fn = NULL);
  BaseQLearnBehavior(spadesalib::AgentCommInterface* pacs,
		     const char* init_qtable_fn,
		     explore_t exp, float alpha_explore, float boltzmann_constant,
		     const char* disable_map_fn = NULL);
  virtual ~BaseQLearnBehavior();

  virtual bool initialize();
  
  void setFinalPolicyFN(std::string fn) { policy_fn = fn; }
  void setFinalQTableFN(std::string fn) { qtable_fn = fn; }
  void setPolicyChangeFN(std::string fn);
  void setStatPolicyTimeLimit(int t) { stat_policy_time_limit = t; }
  void setPolicyEval(std::string fn, int interval, int length);

  //takes over all memory
  // if interval is <= 0, all other params ignored and this is disabled
  void setCompareToActMap(int interval, ActionSubsetMap* pActMap, std::string fn);
  // compares the actions taken to those in an act map
  // note that you must explictly say when you take an action! This is different
  // from the setLastAction function, which deals only with the q-update.
  // if interval is <= 0, all other params ignored and this is disabled
  void setActionCompareToActMap(int interval, ActionSubsetMap* pActMap, std::string fn);
  void actForActionCompare(unsigned state, basic_move_t m);
  
protected:

  void respondSense(SenseType st);
  virtual void respondSenseQLearn(SenseType st) = 0;
  
  //uses last state, last action, getCurrState(), getCurrReward()
  void updateQValue();
  //this should be called at the end of every State sensation;
  //this will update the last state
  void stateSenseUpdate();
  
  QTableFlat qtable;

  explore_t exp_type;
  float alpha_explore;
  float boltzmann_constant;

  ActionSubsetMap disable_map;

  const State& getLastState() const;
  bool isLastValid() const { return last_valid; }
  basic_move_t getLastAction() const { return last_action; }
  //at one point, I wanted to intercept calls to sendMove and stuff, but
  // I think it's better to let the sub classes decide what the last action was
  void setLastAction(basic_move_t m) { last_action = m; }
  
private:
  void writePolicyChangeCount();
  void respondForPolicyEval();
  void handleActmapCompare();
  void handleActmapCompareAction();
  
  bool last_valid;
  State last_state;
  basic_move_t last_action;

  std::string policy_fn;
  std::string qtable_fn;

  std::ofstream os_policy_change;
  int policy_change_count;

  spades::SimTime last_policy_change_time;
  int stat_policy_time_limit;

  std::ofstream os_policy_eval;
  int policy_eval_interval;
  int policy_eval_length;
  bool policy_eval_current;
  //while policy_eval_current, it's the start time, while !policy_eval_current, it's the finish time of
  // the last eval
  spades::SimTime policy_eval_last_time;
  spades::SimTime policy_eval_last_rew_time;
  spades::MeanSummary policy_eval_mean;

  //parameters for the comparing to a fixed action map
  int actmap_compare_interval;
  ActionSubsetMap* actmap_compare_pactmap;
  std::ofstream actmap_compare_os;
  int actmap_compare_last_time;
  
  //parameters for the comparing actions to aa fixed action map
  int actmap_action_compare_interval;
  ActionSubsetMap* actmap_action_compare_pactmap;
  std::ofstream actmap_action_compare_os;
  int actmap_action_compare_last_time;
  spades::MeanSummary actmap_action_compare_summ;
  
  //we don't want to update q for the move from a captured state
  //therefore, we set this flag when we are in a capture state and forbid the update next time
  bool forbid_q_update_next;
};


#endif
