/* This is a base class which does all the parsing of agent messages and
   then calls (virtual) functions to respond to those messages */

#include <strstream>
#include "Behavior.hpp"
#include "Logger.hpp"
#include "PPSharedParam.hpp"

using namespace spadesalib;
using namespace spades;
using namespace std;

Behavior::Behavior(AgentCommInterface* pacs)
  : pacs(pacs),
    visitor(this),
    agentnum(-1), curr_state(), last_reward(0),
    last_reward_time(0), ms_reward()
{
  openRewardFile();
}

Behavior::~Behavior()
{
  //make sure we write out the reward values
  recordReward(true);
}


/******************************************************************************/

//this function is the main message receive loop
//returns the number of messages processed
int
Behavior::behaviorLoop()
{
  int messages_processed = 0;
  FromCSMessage* m;
  
  actionlog(10) << "Beginning main loop" << ende;
  if ( (m=pacs->receiveMessage()) == NULL)
    {
      errorlog << "Did not receive any message on startup" << ende;
      return messages_processed;
    }
  
  FromCSMessage_InitData* m_initdata = dynamic_cast<FromCSMessage_InitData*>(m);
  if (m_initdata == NULL)
    {
      errorlog << "Did not get INIT_DATA first! " << *m << ende;
      return messages_processed;
    }

  messages_processed++;

  pacs->send(new ToCSMessage_InitDone());
  
  //Now we have two cases.
  // * If we received a non-empty init date, this is migration. This means we will not get an
  //   init sensation
  // * It's a normal startup, in which case we should get an init sense first 

  if (m_initdata->getData().getSize() > 0)
    {
      errorlog << "I do not handle migration" << ende;
    }
  else
    {
      actionlog(50) << "Starting up in normal mode " << ende;

      m = pacs->receiveMessage();
      if (m == NULL)
	{
	  errorlog << "Time out/exit while waiting for init sense" << ende;
	  return messages_processed;
	}
      FromCSMessage_Sense* m_initsense = dynamic_cast<FromCSMessage_Sense*>(m);

      if (m_initsense == NULL)
	{
	  errorlog << "I got something weird when I wanted an init sense" << ende;
	  return messages_processed;
	}

      if (!parseInitSense(m_initsense))
	{
	  errorlog << "Error parsing init sense" << ende;
	  return messages_processed;
	}
	  
      pacs->send(new ToCSMessage_DoneThinking());
      
    }

  delete m_initdata;

  if (!initialize())
    {
      errorlog << "Behavior did not initialize after init sense!" << ende;
      return messages_processed;
    }
  
  bool done = false;
  while (!pacs->shouldShutdown() && !done)  
    {
      m = pacs->receiveMessage();
      if (m == NULL)
	{
	  if (!pacs->shouldShutdown())
	    errorlog << "Error in receiving message, exiting" << ende;
	  break;
	}
      
      actionlog(50) << "Got message: " << *m << ende;

      visitor.clearError();
      m->accept(&visitor);
      if (visitor.getError())
	{
	  errorlog << "Error in visiting the message" << ende;
	}

      delete m;
      messages_processed++;
    }

  return messages_processed;
}

void
Behavior::stateSensePostProcess()
{
  recordReward();
  //we now clear the reward to the price we pay for every step
  last_reward = PPSharedParam::instance()->getPerStepReward();
  //same with the received actions and capture signals
  map_action_recv.clear();
  vec_captured_agents.clear();
}

/******************************************************************************/

void
Behavior::sendMove(basic_move_t m)
{
  actionlog(100) << "Behavior: sending move " << m << ende;
  ostrstream out;
  out << "move " << getBasicMoveString(m);
  pacs->send(new ToCSMessage_Act(DataArray(out)));
  //don't unfreeze since DataArray took the memory
}

void
Behavior::sendCoachRequest()
{
  actionlog(90) << "Sending request for action to coach" << ende;
  char* str = "request";
  sendCoachComm(str);
}

void
Behavior::sendAdvice(basic_move_t m)
{
  actionlog(90) << "Sending advice to agent" << ende;
  ostrstream out;
  out << "advice " << m << ends;
  sendComm(out.str());
  out.freeze(false);
}

void
Behavior::sendSAAdvice(const vec_sa_advice_t& vAdvice)
{
  ostrstream out;
  out << "sa_advice ";
  for (vec_sa_advice_t::const_iterator iter = vAdvice.begin();
       iter != vAdvice.end();
       iter++)
    out << *iter << ' ';
  out << ends;
  actionlog(100) << "Behavior: sending sa advice " << out.str() << ende;
  sendComm(out.str());
  out.freeze(false);
}

void
Behavior::sendCoachComm(const char* str)
{
  actionlog(100) << "Behavior: sending coach comm " << str << ende;
  ostrstream out;
  out << "coach_comm " << str;
  pacs->send(new ToCSMessage_Act(DataArray(out)));
  //don't unfreeze since DataArray took the memory
}

void
Behavior::sendComm(const char* str)
{
  actionlog(100) << "Behavior: sending comm " << str << ende;
  ostrstream out;
  out << "comm " << str;
  pacs->send(new ToCSMessage_Act(DataArray(out)));
  //don't unfreeze since DataArray took the memory
}

void
Behavior::sendShutdown()
{
  actionlog(90) << "Sending shutdown request" << ende;
  pacs->send(new ToCSMessage_Act(DataArray("shutdown")));
}


/******************************************************************************/

Behavior::CommType
Behavior::getCommType() const
{
  return comm_type;
}

entityid_t
Behavior::getCommSender() const
{
  return comm_sender;
}

basic_move_t
Behavior::getCommActionAdvice() const
{
  if (comm_type == CT_ActionAdvice)
    {
      return comm_action_advice;
    }
  else
    {
      errorlog << "Shouldn't get comm action advice unless CT_ActionAdvice" << ende;
      return BM_None;
    }
}

const vec_sa_advice_t&
Behavior::getCommSAAdvice() const
{
  if (comm_type != CT_StateActionAdvice)
    errorlog << "Shouldn't getCommSAAdvice if last comm wasn't StateActionAdvice" << ende;
  
  return comm_vec_sa_advice;
}

/******************************************************************************/

bool
Behavior::parseInitSense(FromCSMessage_Sense* m)
{
  istrstream in(m->getData().getData());
  char ch;
  string str;
  
  if (!skip_white_space(in))
    return false;

  in.get(ch);
  if (ch != '(')
    return false;

  in >> str;

  if (str != "init")
    return false;

  in >> agentnum;

  if (!in)
    return false;

  return true;
}

bool
Behavior::parseInform(FromCSMessage_Sense* m)
{
  istrstream in(m->getData().getData());
  char ch;
  string str;

  try
    {
      if (!skip_white_space(in))
	throw ParseException("Initial white space");

      in.get(ch);
      if (ch != '(')
	throw ParseException("No opening paren");

      in >> str;

      if (str == "reward")
	{
	  int rew;
	  in >> rew;
	  if (!in)
	    throw ParseException("Could not read reward number");
	  last_reward += rew;
          actionlog(150) << "Inform: reward of " << rew << ende;
	}
      else if (str == "action")
	{
	  entityid_t eid;
	  basic_move_t m;

	  in >> eid >> m;
	  if (!in)
	    throw ParseException("Could not read action args");

	  map_action_recv[eid] = m;
          actionlog(150) << "Inform: action by " << eid << " of " << m << ende;
	}
      else if (str == "capture")
	{
	  entityid_t eid;

	  in >> eid;
	  if (!in)
	    throw ParseException("Could not read capture args");

	  vec_captured_agents.push_back(eid);
          actionlog(150) << "Inform: capture of " << eid << ende;
	}
      else
	{
	  throw ParseException("Keyword not recognized");
	}

    }
  catch ( ParseException E )
    {
      //All informs could be old sense, though this should actually never happen
      last_reward = 0;
      if (parseSense(m))
	{
	  errorlog << "I don't handle senses becoming informs" << ende;
	  return true;
	}
      errorlog << "Error parsing inform: " << E.getReason() << ende;
      return false;
    }
  
  return true;
}

//example sense:(sense (pred (0 2 2) ) (prey (1 2 2) )  )
//          or  (comm 1 "request")
//          or  (comm 1 "advice E")
//          or  (comm 1 "sa_advice 124 E 133 W 11111 X")
Behavior::SenseType
Behavior::parseSense(FromCSMessage_Sense* m)
{
  istrstream in(m->getData().getData());
  char ch;
  string str;
  State save_state = curr_state;
  SenseType st = ST_None;
  
  try
    {
      if (!skip_white_space(in))
	throw ParseException("Initial whitespace error");

      in.get(ch);
      if (ch != '(')
	throw ParseException("No opening paren");

      in >> str;

      if (str == "sense")
	{
	  in >> curr_state;

	  if (!in)
	    throw ParseException("Could not read state");

	  st = ST_State;
          actionlog(150) << "Sense: state" << ende;
	}
      else if (str == "comm")
	{
	  in >> comm_sender;

	  if (!in)
	    throw ParseException("Could not read comm sender");
	    
	  if (!skip_white_space(in))
	    throw ParseException("Whitespace after sender");
	  
	  in.get(ch);
	  if (ch != '"')
	    throw ParseException("Could not find leading quote");

	  in >> str;

	  /* This is kind of a hack. We find the first quote and delete it
	     If the whoel comm is only one word, the trailing quote may get grabbed also */
	  string::size_type quote_idx = str.find('"');
	  if (quote_idx != string::npos)
	    {
	      str.erase(quote_idx);
	    }

	  
	  if (str == "advice")
	    {
	      comm_type = CT_ActionAdvice;
	      in >> comm_action_advice;
	      if (!in)
		throw ParseException("Could not read action advice");
              actionlog(150) << "Sense: comm advice" << ende;
	    }
	  else if (str == "sa_advice")
	    {
	      comm_type = CT_StateActionAdvice;
	      comm_vec_sa_advice.clear();
	      skip_white_space(in);
	      while (in.peek() != '"')
		{
		  SAAdvice sa;
		  if (in.fail())
		    throw ParseException("SAAdvice: error before reading");
		  in >> sa;
		  if (in.fail())
		    {
		      ostrstream outerr;
		      outerr << "SAAdvice: error reading: " << sa << ends;
		      ParseException exc(outerr.str());
		      outerr.freeze(false);
		      throw exc;
		    }
		      
		  comm_vec_sa_advice.push_back(sa);
		  if (!skip_white_space(in))
		    throw ParseException("SAAdvice: ran out of string!");
		}
              actionlog(150) << "Sense: comm sa_advice" << ende;
	    }
	  else if (str == "request")
	    {
	      comm_type = CT_RequestAction;
              actionlog(150) << "Sense: comm request" << ende;
	    }
	  else
	    {
	      throw ParseException(("Unrecognized keyword in comm: " + str).c_str());
	    }

	  st = ST_Comm;
	}
      else
	{
	  throw ParseException(("Unrecognized keyword for sense type: " + str).c_str());
	}
      
    }
  catch ( ParseException E )
    {
      curr_state = save_state;
      errorlog << "Error parsing sense: " << E.getReason() << ende;
      return ST_None;
    }
      
  return st;
}

/******************************************************************************/

void
Behavior::openRewardFile()
{
  if (PPSharedParam::instance()->getAvgRewardInterval() <= 0)
    return;
  out_avg_reward.open(PPSharedParam::instance()->getAvgRewardFN().c_str());
  if (!out_avg_reward)
    {
      errorlog << "Error opening average reward file '" << PPSharedParam::instance()->getAvgRewardFN() << "'" << ende;
      return;
    }
  out_avg_reward << "# This file was generated by predprey/shared/Behavior" << endl;
  out_avg_reward << "# It contains the average rewards received" << endl;
  out_avg_reward << "# Format: <num_time_steps> <avg_reward>" << endl;
  
}

void
Behavior::recordReward(bool force_output)
{
  if (PPSharedParam::instance()->getAvgRewardInterval() <= 0)
    return;

  SimTime currtime = pacs->getLatestTime();
  
  ms_reward.addPoint((float)last_reward / (float)(currtime - last_reward_time), currtime - last_reward_time);
  last_reward_time = currtime;

  actionlog(220) << "recordReward: " << currtime << ' ' << ms_reward << ende;
  
  if (force_output || ms_reward.getNumPoints() >= PPSharedParam::instance()->getAvgRewardInterval())
    {
      out_avg_reward << "# ending time: " << currtime << endl;
      out_avg_reward << ms_reward.getNumPoints()
		     << '\t' << ms_reward.getMean()
		     << endl;
      ms_reward.reset();
    }
}

  
/******************************************************************************/
void
Behavior::MessageVisitor::visit(FromCSMessage_InitData* m)
{
  errorlog << "Should not get init data here!" << ende;
  error = true;
}

void
Behavior::MessageVisitor::visit(FromCSMessage_Exit* m)
{
  actionlog(50) << "Received an exit" << ende;
  beh->pacs->initiateShutdown();
}

void
Behavior::MessageVisitor::visit(FromCSMessage_Sense* m)
{
  if (spades::startsThinkingCycle(m->getThinking()))
    {
      switch (beh->parseSense(m))
	{
	case ST_State:
	  actionlog(100) << "Behavior: responding to state" << ende;
	  beh->respondSense(ST_State);
	  beh->stateSensePostProcess();
	  break;
	case ST_Comm:
	  actionlog(100) << "Behavior: responding to comm" << ende;
	  beh->respondSense(ST_Comm);
	  break;
	case ST_None:
	  errorlog << "Parsing on sense failed" << ende;
	  error = true;
	  break;
	case ST_TimeNotify:
	  errorlog << "Should not get time notifies here!" << ende;
	  break;
	}
      beh->pacs->send(new ToCSMessage_DoneThinking());
    }
  else
    {
      if (beh->parseInform(m))
	{
	  actionlog(100) << "Behavior: notifying of inform" << ende;
	  beh->notifyInform();
	}
      else
	{
	  errorlog << "Parsing of inform failed" << ende;
	}
    }
}

void
Behavior::MessageVisitor::visit(FromCSMessage_TimeNotify* m)
{
  if (spades::startsThinkingCycle(m->getThinking()))
    {
      actionlog(100) << "Behavior: responding to time notify" << ende;
      beh->respondSense(ST_TimeNotify);
      beh->pacs->send(new ToCSMessage_DoneThinking());
    }
  else
    {
      actionlog(100) << "Behavior: notifying old time notify" << ende;
      beh->notifyOldTimeNotify();
    }
}

void
Behavior::MessageVisitor::visit(FromCSMessage_MigrationRequest* m)
{
  errorlog << "Don't handle mimgration" << ende;
  error = true;
}

void
Behavior::MessageVisitor::visit(FromCSMessage_Error* m)
{
  errorlog << "Got an error from commserver: " << m->getString() << ende;
  error = true;
}

void
Behavior::MessageVisitor::visit(FromCSMessage_ThinkTime* m)
{
  actionlog(200) << "Got think time of " << m->getTime() << ende;
}
