/* This is a base class which does all the parsing of agent messages and
   then calls (virtual) functions to respond to those messages */

#ifndef _BEHAVIOR_H_
#define _BEHAVIOR_H_

#include <map>
#include <fstream>
#include "AgentCommInterface.hpp"
#include "PPSharedTypes.hpp"
#include "State.hpp"
#include "SAAdvice.hpp"
#include "utility.hpp"

class Behavior 
{
public:
  Behavior(spadesalib::AgentCommInterface* pacs);
  virtual ~Behavior();

  //this function is the main message receive loop
  //returns the number of messages processed
  int behaviorLoop();

  enum SenseType {
    ST_None,
    ST_TimeNotify,
    ST_State,
    ST_Comm
  };
  
  enum CommType {
    CT_None,
    CT_RequestAction,
    CT_ActionAdvice, //an action to do right now
    CT_StateActionAdvice // a set of state action advice pairs
  };

private:
  class MessageVisitor
    : public spadesalib::FromCSMessage::Visitor
  {
  public:
    MessageVisitor(Behavior* b) : spadesalib::FromCSMessage::Visitor(), beh(b), error(false) {}
    ~MessageVisitor() {}

    void visit(spadesalib::FromCSMessage_InitData* m);
    void visit(spadesalib::FromCSMessage_Exit* m);
    void visit(spadesalib::FromCSMessage_Sense* m);
    void visit(spadesalib::FromCSMessage_TimeNotify* m);
    void visit(spadesalib::FromCSMessage_MigrationRequest* m);
    void visit(spadesalib::FromCSMessage_Error* m);
    void visit(spadesalib::FromCSMessage_ThinkTime* m);

    bool getError() const { return error; }
    void clearError() { error = false; }
    
  private:
    Behavior* beh;
    bool error;
  };

protected:
  friend class MessageVisitor;
  
  //Now the virtual functions that behaviors have to define
  //this is going to need an argument about the type of sense
  virtual void respondSense(SenseType st) = 0;
  virtual void notifyInform() {} 
  virtual void notifyOldTimeNotify() {} 
  // this is called after init sense parsing
  virtual bool initialize() = 0;

  //These are functions the behavior should call to do stuff
  void sendMove(basic_move_t m);
  void sendCoachRequest(); //sends to the coach
  void sendAdvice(basic_move_t m); //sends from the coach
  void sendSAAdvice(const vec_sa_advice_t& vAdvice); //sends from the coach
  void sendShutdown(); //requests an exit 

  //These are functions that the subclass can call to get info
  int   getAgentNum() const { return agentnum; }
  const State& getCurrState() const { return curr_state; }
  int   getLastReward() const { return last_reward; }

  //information about the actions we were notified about
  typedef std::map<entityid_t,basic_move_t> map_action_recv_t;
  const map_action_recv_t& getMapActionRecv() const { return map_action_recv; }

  //information about captures
  typedef std::vector<entityid_t> vec_captured_agents_t;
  const vec_captured_agents_t& getVecCapturedAgents() const { return vec_captured_agents; }
  
  //These are only use for communication
  //SMURF: This is a big hack right now, there needs to be a better way to do this
  CommType     getCommType() const;
  entityid_t   getCommSender() const;
  basic_move_t getCommActionAdvice() const;

  const vec_sa_advice_t& getCommSAAdvice() const;
  
  spadesalib::AgentCommInterface* pacs;

  
protected:
  //This stuff is really for the message visitor, the subclasses shouldn't use it
  
  class ParseException
  {
  public:
    ParseException(const char* reason) : reason(reason) {}
    ~ParseException() {}

    std::string getReason() const { return reason; }
  private:
    std::string reason;
  };
  
  bool parseInitSense(spadesalib::FromCSMessage_Sense* m);
  bool parseInform(spadesalib::FromCSMessage_Sense* m);
  SenseType parseSense(spadesalib::FromCSMessage_Sense* m);

  void stateSensePostProcess();
  
private:

  void openRewardFile();
  void recordReward(bool force_output = false);

  //there are public versions for the sub classes
  void sendCoachComm(const char* str);
  void sendComm(const char* str);

  MessageVisitor visitor;
  
  int agentnum;
  State curr_state;
  int last_reward;

  map_action_recv_t map_action_recv;
  vec_captured_agents_t vec_captured_agents;
  
  CommType     comm_type;
  entityid_t   comm_sender;
  basic_move_t comm_action_advice;
  vec_sa_advice_t comm_vec_sa_advice;
  
  std::ofstream out_avg_reward;
  spades::SimTime last_reward_time;
  spades::MeanSummary ms_reward;
  
};



#endif
