#include <strstream>

#include "EngineParam.hpp"
#include "SimEngine.hpp"
#include "EndSimulationEvent.hpp"

#include "PredPreyWorld.hpp"
#include "random.hpp"
#include "Logger.hpp"
#include "MoveActEvent.hpp"
#include "CommActEvent.hpp"
#include "ReqShutdownActEvent.hpp"
#include "CreateFullSenseEvent.hpp"
#include "CreateRewardInformEvent.hpp"
#include "CreateCaptureInformEvent.hpp"
#include "RandomPlaceEvent.hpp"
#include "InitSenseEvent.hpp"

using namespace std;
using namespace spades;

/*************************************************************************************/

PredPreyWorld::PredPreyWorld ()
  : WorldModel (), pSimEngine(NULL), next_entity_id(0),
    map_pred(), map_prey(), pPredCoach(NULL), pPreyCoach(NULL)
{
}

PredPreyWorld::~PredPreyWorld ()
{
  if (pPredCoach) delete pPredCoach;
  if (pPreyCoach) delete pPreyCoach;
}


EngineParam *
PredPreyWorld::parseParameters (const int argc, const char *const *argv)
{
  PredPreyParam* p = new PredPreyParam (argc, argv);

  //cout << p->getAgentTypeDB() << endl;
  
  return p;
}


/* Most of the real initialization/finalization work should be done here */
bool
PredPreyWorld::initialize (SimEngine * pSE)
{
  actionlog (10) << "PredPreyWorld: Initializing" << ende;

  pSimEngine = pSE;

  pSimEngine->enqueueEvent(new EndSimulationEvent(PredPreyParam::instance()->getSimulationLength()));

  pred_type = pSE->getAgentTypeDB()->getAgentType(PredPreyParam::instance()->getPredName());
  if (pred_type == pSE->getAgentTypeDB()->nullIterator())
    errorlog << "I could not find an agent of the pred type: " << PredPreyParam::instance()->getPredName() << ende;

  prey_type = pSE->getAgentTypeDB()->getAgentType(PredPreyParam::instance()->getPreyName());
  if (prey_type == pSE->getAgentTypeDB()->nullIterator())
    errorlog << "I could not find an agent of the prey type: " << PredPreyParam::instance()->getPreyName() << ende;

  if (PredPreyParam::instance()->getUsePredCoach())
    {
      pred_coach_type = pSE->getAgentTypeDB()->getAgentType(PredPreyParam::instance()->getPredCoachName());
      if (pred_coach_type == pSE->getAgentTypeDB()->nullIterator())
	errorlog << "I could not find an agent of the pred coach type: " << PredPreyParam::instance()->getPredCoachName() << ende;
    }
  
  if (PredPreyParam::instance()->getUsePreyCoach())
    {
      prey_coach_type = pSE->getAgentTypeDB()->getAgentType(PredPreyParam::instance()->getPreyCoachName());
      if (prey_coach_type == pSE->getAgentTypeDB()->nullIterator())
	errorlog << "I could not find an agent of the prey coach type: " << PredPreyParam::instance()->getPreyCoachName() << ende;
    }

  return true;
}

bool
PredPreyWorld::finalize ()
{
  actionlog (10) << "PredPreyWorld: Finalizing" << ende;
  return true;
}

void
PredPreyWorld::pauseModeCallback ()
{
  if (pSimEngine->getNumCommServers () < PredPreyParam::instance()->getNumCommServersWanted())
    {
      actionlog(100) << "Not enough comm servers, waiting "
                     << pSimEngine->getNumCommServers() << "<" << PredPreyParam::instance()->getNumCommServersWanted()
                     << ende;
      return;
    }

  /* the world count is only for agents that have successfully connected
     the engine count is for all agent in any stage of startup */
  int engine_agent_count = pSimEngine->getNumAgents();
  int world_agent_count = map_pred.size() + map_prey.size() + (pPredCoach != NULL) + (pPreyCoach != NULL);

  if (engine_agent_count == world_agent_count)
    {
      //look into starting up more agents!
      if (world_agent_count == PredPreyParam::instance()->getNumPred() + PredPreyParam::instance()->getNumPrey() +
	  PredPreyParam::instance()->getUsePredCoach() + PredPreyParam::instance()->getUsePreyCoach())
	{
	  //We have all the agents we want. Let's go!
	  actionlog(100) << "All agents started" << ende;
	  if (pSimEngine->areAllAgentsInitialized())
	    pSimEngine->changeSimulationMode(SM_RunNormal);
	  else
	    actionlog(100) << "Waiting for agents to initialize..." << ende;
	}
      else
	{
	  actionlog(100) << "Starting more agents" << ende;
	  
	  AgentID a;

	  if (map_pred.size() < (unsigned)PredPreyParam::instance()->getNumPred())
	    {
	      for (int i=map_pred.size(); i<PredPreyParam::instance()->getNumPred(); i++)
		{
		  actionlog(110) << "Starting a pred" << ende;
		  if ( (a=pSimEngine->startNewAgent (pred_type)) == AGENTID_INVALID)
		    errorlog << "Could not start a pred agent" << ende;
		}
	    }

	  else if (map_prey.size() < (unsigned)PredPreyParam::instance()->getNumPrey())
	    {
	      for (int i=map_prey.size(); i<PredPreyParam::instance()->getNumPrey(); i++)
		{
		  actionlog(110) << "Starting a prey" << ende;
		  if ( (a=pSimEngine->startNewAgent (prey_type)) == AGENTID_INVALID)
		    errorlog << "Could not start a prey agent" << ende;
		}
	    }

	  else if (PredPreyParam::instance()->getUsePredCoach() && pPredCoach == NULL)
	    {
	      a = pSimEngine->startNewAgent (pred_coach_type);
	      if (a == AGENTID_INVALID)
		errorlog << "Could not start pred coach" << ende;
	    }
	  
	  else if (PredPreyParam::instance()->getUsePreyCoach() && pPreyCoach == NULL)
	    {
	      a = pSimEngine->startNewAgent (prey_coach_type);
	      if (a == AGENTID_INVALID)
		errorlog << "Could not start prey coach" << ende;
	    }
      
	}
    }
}


SimTime
PredPreyWorld::simToTime (SimTime time_curr, SimTime time_desired)
{
  actionlog(100) << "Advancing time from " << time_curr << " to " << time_desired << ende;
  SimTime t;
  //we imgaine that all this movement happens at the very very begining of the time step
  // therfore, we always have to advance by at least 1 step, since at time_curr, everything has
  // already happened
  for (t = time_curr + 1;
       true/* we do the real break check below to make sure t doesn't advance past time_desired */;
       t++)
    {
      if (t % PredPreyParam::instance()->getDiscreteStepSize() == 0)
	{
	  actionlog(60) << "Advancing world state at " << t << ende;

 	  for (map_pred_t::iterator iter = map_pred.begin();
	       iter != map_pred.end();
	       ++iter)
	    {
	      iter->second.applyMove();
	    }

	  if (!PredPreyParam::instance()->getFixPreyLocations())
	    {
	      for (map_prey_t::iterator iter = map_prey.begin();
		   iter != map_prey.end();
		   ++iter)
		{
		  iter->second.applyMove();
		}
	    }
	  
	  int count = checkReward(t);
	  actionlog(120) << "Reward signals sent: " << count << ende;
	  /* If rewards were queued up, we now have events in the queue, so we need to stop here */
	  if (count != 0)
	    break;
	}

      if (t == time_desired)
	break;
    }

  logPositions(210);
  
  return t;
}


spades::DataArray
PredPreyWorld::getMonitorHeaderInfo ()
{
  ostrstream out;
  out << PredPreyParam::instance()->getXSize() << ' ' << PredPreyParam::instance()->getYSize() << endl;
  return DataArray(out);
}

spades::DataArray
PredPreyWorld::getMonitorInfo (spades::SimTime time)
{      	
  ostrstream out;
  out << time << ": ";
  out << "(pred ";
  for (map_pred_t::iterator iter = map_pred.begin();
       iter != map_pred.end();
       ++iter)
    {
      iter->second.printCompact(out);
    }
  out << ") ";
  
  out << "(prey ";
  for (map_prey_t::iterator iter = map_prey.begin();
       iter != map_prey.end();
       ++iter)
    {
      iter->second.printCompact(out);
    }
  out << ") ";

  out << endl;
  
  return DataArray(out);
}

void
PredPreyWorld::parseMonitorMessage (const char* data, unsigned datalen)
{
  errorlog << "I don't understand anything from the monitor" << ende;
}



bool
PredPreyWorld::agentConnect (AgentID agent, AgentTypeDB::AgentTypeConstIterator at)
{
  int entity_id = -1;
  if (at == pred_type)
    {
      Entity e;
      entity_id = next_entity_id++;
      e.setID(entity_id);
      e.setPosRandom();
      map_pred[agent] = e;

    }
  else if (at == prey_type)
    {
      Entity e;
      entity_id = next_entity_id++;
      e.setID(entity_id);
      if (PredPreyParam::instance()->getFixPreyLocations())
	e.setLoc(Location(0,0));
      else
	e.setPosRandom();
      map_prey[agent] = e;
    }
  else if (at == pred_coach_type)
    {
      if (pPredCoach != NULL)
	{
	  errorlog << "Trying to connect multiple pred coaches!" << ende;
	}
      entity_id = PRED_COACH_ID;
      pPredCoach = new CoachEntity(agent, entity_id);
    }
  else if (at == prey_coach_type)
    {
      if (pPreyCoach != NULL)
	{
	  errorlog << "Trying to connect multiple prey coaches!" << ende;
	}
      entity_id = PREY_COACH_ID;
      pPreyCoach = new CoachEntity(agent, entity_id);
    }  
  else
    {
      errorlog << "I don't understand what the agent type is!: "
	       << pSimEngine->getAgentTypeDB()->deref(at) << ende;
      return false;
    }
  
  //create an initial sensation
  InitSenseEvent* pE_init =
    new InitSenseEvent(pSimEngine->getSimulationTime(),
		       agent, entity_id);
  pSimEngine->enqueueEvent(pE_init);

  //Create a new sensation event
  CreateFullSenseEvent* pE_csense =
    new CreateFullSenseEvent(pSimEngine->getSimulationTime() + 1, agent);
  pSimEngine->enqueueEvent(pE_csense);

  return true;
}

bool
PredPreyWorld::agentDisappear (AgentID agent, spades::AgentLostReason reason)
{
  //actionlog(50) << "PredPrey removing agent " << agent << ende;
  errorlog << "PredPrey removing agent " << agent << ende;

  map_pred.erase(agent);
  map_prey.erase(agent);
  if (pPredCoach && pPredCoach->getAgentID() == agent)
    {
      delete pPredCoach;
      pPredCoach = NULL;
    }
  if (pPreyCoach && pPreyCoach->getAgentID() == agent)
    {
      delete pPreyCoach;
      pPreyCoach = NULL;
    }

  pSimEngine->enqueueEvent(new EndSimulationEvent(pSimEngine->getSimulationTime() + 10));
  
  return true;
}

ActEvent*
PredPreyWorld::parseAct(SimTime t, AgentID a, const char* data, unsigned datalen) const
{
  ActEvent* pE = NULL;
  istrstream in(data, datalen);
  string str;
  in >> str;
  if (str == "move")
    {
      actionlog(100) << "Received move from agent " << a << ende;

      basic_move_t m;
      in >> m;
      if (!in)
	{
	  std::string s(data, datalen);
	  errorlog << "Did not understand move from agent: " << s << ende;
	}
      else
	{
	  pE = new MoveActEvent(t, a, m);
	}
    }
  else if (str == "coach_comm")
    {
      actionlog(100) << "Received coach_comm from agent " << a << ende;
      
      pp_agent_type_t atype;
      entityid_t eid = getEntityIDFromAgentID(a, &atype);
      CommActEvent* pEComm = createCommEvent(t, a, eid, in);
      
      //set the receivers
      switch (atype)
	{
	case PPAT_Pred:
	  if (pPredCoach)
	    pEComm->addReceiver(pPredCoach->getAgentID());
	  else
	    actionlog(110) << "coach comm, but no pred coach" << ende;
	  break;
	case PPAT_Prey:
	  if (pPreyCoach)
	    pEComm->addReceiver(pPreyCoach->getAgentID());
	  else
	    actionlog(110) << "coach comm, but no prey coach" << ende;
	  break;
	case PPAT_None:
	case PPAT_CoachPred:
	case PPAT_CoachPrey:
	  errorlog << "Bad atype for coach_comm: " << atype << ende;
	  delete pEComm;
	  pEComm = NULL;
	  break;
	default:
	  errorlog << "I don't understand atype: " << atype << ende;
	  delete pEComm;
	  pEComm = NULL;
	  break;
	}
      pE = pEComm;
    }
  else if (str == "comm")
    {
      actionlog(100) << "Received comm from agent " << a << ende;

      pp_agent_type_t atype;
      entityid_t eid = getEntityIDFromAgentID(a, &atype);
      CommActEvent* pEComm = createCommEvent(t, a, eid, in);
      
      //set the receivers
      switch (atype)
	{
	case PPAT_CoachPred:
	  for (map_pred_t::const_iterator iter = map_pred.begin();
	       iter != map_pred.end();
	       iter++)
	    pEComm->addReceiver(iter->first);
	  break;
	case PPAT_CoachPrey:
	  for (map_prey_t::const_iterator iter = map_prey.begin();
	       iter != map_prey.end();
	       iter++)
	    pEComm->addReceiver(iter->first);
	  break;
	case PPAT_None:
	case PPAT_Pred:
	case PPAT_Prey:
	  errorlog << "Bad atype for comm: " << atype << ende;
	  delete pEComm;
	  pEComm = NULL;
	  break;
	default:
	  errorlog << "I don't understand atype: " << atype << ende;
	  delete pEComm;
	  pEComm = NULL;
	  break;
	}
      pE = pEComm;
    }
  else if (str == "shutdown")
    {
      pE = new ReqShutdownActEvent(t, a);
    }
  else
    {
      errorlog << "Did not understand action from agent: " << (char*)data << ende;
      return NULL;
    }

  return pE;
}

CommActEvent*
PredPreyWorld::createCommEvent(spades::SimTime t, spades::AgentID a, entityid_t eid,
			       istream& in) const
{
  skip_white_space(in);
  //SMURF: should really handle the length better
  const int max_len = 1024;
  char mess[max_len];
  int len = in.rdbuf()->sgetn(mess, max_len);
  if (len == max_len)
    errorlog << "communication message is too long: " << len << ende;
  mess[len] = 0;

  return new CommActEvent(t, a, eid, mess, len);
}


SimTime
PredPreyWorld::getSenseInterval()
{
  return PredPreyParam::instance()->getDiscreteStepSize();
}

int
PredPreyWorld::checkReward(spades::SimTime time)
{
  int count = 0;

  //This is a simple scheme where all predators get a fixed reward for every prey
  // that has a predator right on top of it
  // we'll do this an inefficient way for now because it doesn't matter
  for (map_prey_t::iterator iter_prey = map_prey.begin();
       iter_prey != map_prey.end();
       ++iter_prey)
    {
      for (map_pred_t::iterator iter_pred = map_pred.begin();
	   iter_pred != map_pred.end();
	   ++iter_pred)
	{
	  if (iter_prey->second.isSameLocation(iter_pred->second))
	    {
	      for (map_pred_t::iterator iter_pred2 = map_pred.begin();
		   iter_pred2 != map_pred.end();
		   ++iter_pred2)
		{
		  pSimEngine->enqueueEvent(new CreateRewardInformEvent(time, iter_pred2->first,
								       PredPreyParam::instance()->getReward()));
		  pSimEngine->enqueueEvent(new CreateCaptureInformEvent(time, iter_pred2->first,
								       iter_prey->second.getID()));
		  count++;
		}
	      if (pPredCoach)
		{
		  pSimEngine->enqueueEvent(new CreateRewardInformEvent(time, pPredCoach->getAgentID(),
								       PredPreyParam::instance()->getReward()));
		  pSimEngine->enqueueEvent(new CreateCaptureInformEvent(time, pPredCoach->getAgentID(),
								       iter_prey->second.getID()));
		  count++;
		}
	      
	      pSimEngine->enqueueEvent(new RandomPlaceEvent(time+PredPreyParam::instance()->getDiscreteStepSize() - 1,
							    iter_pred->second));
	      if (!PredPreyParam::instance()->getFixPreyLocations())
		pSimEngine->enqueueEvent(new RandomPlaceEvent(time+PredPreyParam::instance()->getDiscreteStepSize() - 1,
							      iter_prey->second));
	      break; // no need to check other predators
	    }
	}
    }

  return count;
}

void
PredPreyWorld::logPositions(int level)
{
  actionlog(level) << "pred: ";
  for (map_pred_t::iterator iter = map_pred.begin();
       iter != map_pred.end();
       ++iter)
    {
      actionlog(level) << iter->second;
    }
  actionlog(level) << ende;
  
  actionlog(level) << "prey: ";
  for (map_prey_t::iterator iter = map_prey.begin();
       iter != map_prey.end();
       ++iter)
    {
      actionlog(level) << iter->second;
    }
  actionlog(level) << ende;
}

entityid_t
PredPreyWorld::getEntityIDFromAgentID(spades::AgentID a, pp_agent_type_t *ptype) const
{
  if (ptype) *ptype = PPAT_None;

  if (pPredCoach && pPredCoach->getAgentID() == a)
    {
      if (ptype) *ptype = PPAT_CoachPred;
      return pPredCoach->getEntityID();
    }
  
  if (pPreyCoach && pPreyCoach->getAgentID() == a)
    {
      if (ptype) *ptype = PPAT_CoachPrey;
      return pPreyCoach->getEntityID();
    }

  map_pred_t::const_iterator prediter = map_pred.find(a);
  if (prediter != map_pred.end())
    {
      if (ptype) *ptype = PPAT_Pred;
      return prediter->second.getID();
    }
  
  map_prey_t::const_iterator preyiter = map_prey.find(a);
  if (preyiter != map_prey.end())
    {
      if (ptype) *ptype = PPAT_Prey;
      return preyiter->second.getID();
    }

  //SMURF: we don't really have an error return value!
  errorlog << "Tried to identify entity from agent id I know nothing about! " << a << ende;
  return 0;
}

